#!/usr/bin/env python

from collections import defaultdict
import glob
import numpy as np
import pandas
import xarray

import smartcarb.io
import smartcarb.misc

from . import timeprofiles as tp




def filter_results(ts, is_mce=False, nmax=None):
    """
    Remove bad fits of flux inversion.
    """

    # analytical inversion
    if 'CO2 emissions (Mt/yr)' in ts or 'NO2 emissions (kt/yr)' in ts:

        tracer = 'CO2' if 'CO2 emissions (Mt/yr)' in ts else 'NO2'

        if tracer == 'CO2':
            good = np.isfinite(ts['CO2 emissions (Mt/yr)'])
            good = ts['CO2 emissions (Mt/yr)'] > 0.0
            good &= ts['CO2 uncertainty (Mt/yr)'] < 10.0

            return (
                ts['CO2 emissions (Mt/yr)'][good],
                ts['CO2 uncertainty (Mt/yr)'][good],
                ts.lon_eq[good],
                good
            )
        else:
            good = np.isfinite(ts['NO2 emissions (kt/yr)'])
            good &= ts['NO2 emissions (kt/yr)'] > 0
            good &= ts['NO2 uncertainty (kt/yr)'] < 10.0

            return (
                ts['NO2 emissions (kt/yr)'][good],
                ts['NO2 uncertainty (kt/yr)'][good],
                ts.lon_eq[good],
                good
            )

    else: # mass-balance approach TODO: remove this

        if is_mce:
            ts, ts_std, ts_sizes = dplume.io.average_results(ts, add_size=True)

        good = np.isfinite(ts['est. CO2 emissions (Mt/yr)'])

        if nmax is not None:
            if 'plume size' in ts:
                good &= ts['plume size'] > nmax
            else:
                print('Warning: number of pixels nmax not used.')

        if 'failed' in ts:
            good &= ts.failed == 0


        ts = ts[good]
        est = ts['est. CO2 emissions (Mt/yr)']
        lon_eqs = ts.lon_eq

        if is_mce:
            est_std = ts_std['est. CO2 emissions (Mt/yr)']
        else:
            est_std = ts['est. CO2 emissions uncertainty (Mt/yr)']

        return est, est_std, lon_eqs, goods


def read_results(filename, nsat=1):
    ts = pandas.read_csv(filename, delimiter='\t', index_col=0,
                         parse_dates=True, na_values='%10.3f' % np.nan)
    ts = smartcarb.misc.filter_dataset(ts, nsat)

    return ts


def read_analytical_inversion(filename, nsat=1, use_constant=False, origin='Berlin', nmax=None,
            tracer='CO2', is_mce=False, fit_cycle=True, n_knots=4):
    """
    Read and prepare time series of estimated and true emissions as well as the
    fitted seasonal cycle.
    """
    ts = read_results(filename, nsat)
    lon_eqs_in_constellation = smartcarb.misc.lon_eq_by_nsat(nsat)

    em_true = smartcarb.io.read_emissions(origin=origin, time='raw',
                                          use_constant=use_constant,
                                          tracer=tracer)
    if tracer == 'NO2':
        em_true *= 1e3
        units = 'kt NO2 yr-1'
    else:
        units = 'Mt CO2 yr-1'

    # remove bad values
    est, est_std, lon_eq, goods = filter_results(ts, is_mce=is_mce, nmax=nmax)

    # create xarray dataset with all data
    d = xarray.Dataset()

    d['overpass'] = xarray.DataArray(est.index, dims='overpass')
    d['est_emissions'] = xarray.DataArray(est, dims='overpass')
    d['est_emissions_std'] = xarray.DataArray(est_std, dims='overpass')
    d['lon_eq'] = xarray.DataArray(lon_eq, dims='overpass')


    d['time'] = xarray.DataArray(em_true.index, dims='time')
    d['true_emissions'] = xarray.DataArray(em_true.values, dims='time')

    if fit_cycle:
        # fit seasonal cycle
        fit, func, times, cycle, chi2 = tp.fit(d['overpass'], d['est_emissions'],
                                               d['est_emissions_std'], n_knots=n_knots)

        d['fitted_cycle'] = xarray.DataArray(cycle, dims='time')
        d['fitted_cycle'].attrs['annual_mean'] = func.integrate(fit.x)
        d['fitted_cycle'].attrs['units'] = units

    return d





def read_mass_balance(tracer, which='combined', method='sub-areas'):


    bad_times = [
        #np.datetime64('2015-01-24 11:00'),
        np.datetime64('2015-07-03 11:00'), # janschwalde in plume
        np.datetime64('2015-10-01 11:00'), # janschwalde in plume
    ]

    # add true emissions
    em_true = smartcarb.io.read_emissions(origin='Berlin', time='raw',
                                          use_constant=False,
                                          tracer=tracer)

    if tracer == 'NO2':
        em_true *= 1e3 # Mt yr-1 -> kt yr-1

    dataset = xarray.Dataset()
    dataset['time'] = xarray.DataArray(em_true.index, dims='time')
    dataset['true_emissions'] = xarray.DataArray(em_true.values, dims='time')

    # add results from mass-balance approach
    data = defaultdict(list)
    filenames = sorted(glob.glob(
        "/output/SMARTCARB/output/mass-balance/CO2-medium_and_NO2-high/*_%s_%s-results.nc"
        % (method, tracer)
    ))

    varname = {
        'model':      'flux_from_model_tracer_using_true_wind_fit',
        'combined':   'flux_from_%s_minus_estimated_background_using_mean_wind_fit' % tracer,
        'background': 'flux_from_%s_noisefree_minus_estimated_background_using_effective_wind_fit' % tracer,
        'wind':       'flux_from_%s_true_signal_using_mean_wind_fit' % tracer,
        'retrieval':  'flux_from_%s_minus_true_background_using_effective_wind_fit' % tracer,
        'method':     'flux_from_%s_true_signal_using_effective_wind_fit' % tracer
    }[which]

    for filename in filenames:
        d = xarray.open_dataset(filename)

        if pandas.Timestamp(d.attrs['time']) in bad_times:
            print('Skip bad overpass')
            continue

        if d.attrs['wind_speed_500m'] < 2.0:
            print('Skip wind speed below 2 m/s')
            continue


        data['overpass'].append( pandas.Timestamp(d.attrs['time']) )
        data['orbit'].append( d.attrs['orbit'] )
        data['lon_eq'].append( d.attrs['lon_eq'] )


        if d[varname].emissions < 0 or d[varname].emissions > 100:
            print('Est. emission of %.1f set to nan. ' % d[varname].emissions)
            data['est_emissions'].append(np.nan)
            data['est_emissions_std'].append(np.nan)
        else:
            data['est_emissions'].append(d[varname].emissions)
            data['est_emissions_std'].append(d[varname].emissions_std)

        # wind speed and direction
        data['wind_speed_500m'].append( d.attrs['wind_speed_500m'] )
        data['wind_direction_500m'].append( d.attrs['wind_direction_500m'] )
        data['wind_speed_eff'].append( d.attrs['wind_speed_eff'] )

        d.close()

    for key, values in data.items():
        dataset[key] = xarray.DataArray(values, dims='overpass')

    return dataset
